from sync.executing import SafeCommandExecutor
from sync.links import LinkSynchronizer

from youtrack.sync.issues import AsymmetricIssueMerger

query_time_format = '%m-%dT%H:%M:%S'
batch = 100
tag = "sync"
master_sync_field_name = 'Sync with'
empty_field_text = 'No sync with'

def get_formatted_for_query(_datetime):
    return _datetime.strftime(query_time_format)

def get_advanced_query(query, _last_run, _current_run):
    return query + ' updated: ' + get_formatted_for_query(_last_run) + " .. " + get_formatted_for_query(_current_run)

class YouTrackSynchronizer(object):
    def __init__(self, master, slave, logger, issue_binder, project_id, fields_to_sync, query, last_run=None, current_run=None):
        self.slave = None
        self.master = master
        self.slave = slave
        self.logger = logger
        self.master_executor = SafeCommandExecutor(master, logger)
        self.slave_executor = SafeCommandExecutor(slave, logger)
        self.issue_binder = issue_binder
        self.query = query
        self.last_run = last_run
        self.current_run = current_run
        self.project_id = project_id
        self.link_synchronizer = LinkSynchronizer(self.master_executor, self.slave_executor, self.issue_binder)
        self.issue_synchronizer = AsymmetricIssueMerger(master, slave, self.master_executor, self.slave_executor, self.issue_binder, self.link_synchronizer, fields_to_sync, last_run, current_run, project_id)

    def setDebugMode(self, on):
        self.master_executor.setDebugMode(on)
        self.slave_executor.setDebugMode(on)

    def sync(self):
        #0. create if not existed and attach synchronization field
        self._create_and_attach_sync_field(self.slave, self.project_id, master_sync_field_name)

        #1. synchronize sync-issues in slave which have no synchronized clone in master
        imported_slave_ids_set = self._apply_to_issues(self._get_tagged_only_in_slave,
            self._import_to_master,
            log_header='[Sync, Importing new issues from slave to master]')

        #2. synchronize sync-issues in master which have no synchronized clone in slave
        sync_set = set(self.issue_binder.s_to_m.values())
        imported_master_ids_set = self._apply_to_issues(self._get_tagged_in_master,
            self._import_to_slave,
            excluded_ids=sync_set,
            log_header='[Sync, Importing new issues from master to slave]')

        #3. synchronize sync-issues updated in slave which have synchronized clone in master
        updated_slave_ids_set = self._apply_to_issues(self._get_updated_in_slave_from_last_run,
            self._sync_to_master,
            excluded_ids=imported_slave_ids_set,
            log_header='[Sync, Merging sync issues updated in slave]')

        #4. synchronize sync-issues updated in master which have synchronized clone in slave (if clone hasn't been updated)
        updated_master_ids_set = self._slave_ids_set_to_sync_ids_set(updated_slave_ids_set) | imported_master_ids_set
        self._apply_to_issues(self._get_updated_in_master_from_last_run,
            self._sync_to_slave,
            excluded_ids=updated_master_ids_set,
            log_header='[Sync, Merging sync issues updated in master and unchanged in slave]')

        #5. synchronize links
        self.link_synchronizer.syncCollectedLinks()

    def syncAfterImport(self):
        self._create_and_attach_sync_field(self.slave, self.project_id, master_sync_field_name)
        start = 0
        issues = self.slave.getIssues(self.project_id, '', start, batch)
        while len(issues):
            for issue in issues:
                issue_id = issue.id
                issue_number = issue_id.rpartition('-')[2]
                self._mark_issues_as_sync(issue_number, issue_id, issue_id)
            start += batch
            issues = self.slave.getIssues(self.project_id, '', start, batch)

    def _slave_ids_set_to_sync_ids_set(self, ids):
        return set([self.issue_binder.slaveIssueIdToMasterIssueId(id) for id in ids])

    def _import_to_master(self, slave_issue):
        master_issue_id = self.issue_synchronizer.clone_issue_to_master(slave_issue)
        if master_issue_id:
            self._mark_issues_as_sync(master_issue_id.rpartition('-')[2], master_issue_id, slave_issue.id)

    def _import_to_slave(self, master_issue):
        slave_issue_id = self.issue_synchronizer.clone_issue_to_slave(master_issue)
        if slave_issue_id:
            self._mark_issues_as_sync(master_issue.numberInProject, master_issue.id, slave_issue_id)

    def _sync_to_master(self, slave_issue):
        self.issue_synchronizer.sync(None, slave_issue.id)

    def _sync_to_slave(self, master_issue):
        self.issue_synchronizer.sync(master_issue.id, None)

    def _apply_to_issues(self, issues_getter, action, excluded_ids=None, log_header=''):
        if not issues_getter or not action: return
        start = 0
        print(log_header + ' started...')
        issues = issues_getter(start, batch)
        processed_issue_ids_set = set([])
        while len(issues):
            for issue in issues:
                sync_id = str(issue.id)
                if not (excluded_ids and sync_id in excluded_ids):
                    action(issue)
                    processed_issue_ids_set.add(sync_id)
            print(log_header + ' processed ' + str(start + len(issues)) + ' issues')
            start += batch
            issues = issues_getter(start, batch)
        print(log_header + ' action applied to ' + str(len(processed_issue_ids_set)) + ' issues')
        return processed_issue_ids_set

    def _get_tagged_only_in_slave(self, start, batch):
        rq = self.query + ' ' + master_sync_field_name + ':  {' + empty_field_text + '}'
        return self.slave.getIssues(self.project_id, rq, start, batch)

    def _get_tagged_in_master(self, start, batch):
        rq = self.query
        return self.master.getIssues(self.project_id, rq, start, batch)

    def _get_updated_in_slave_from_last_run(self, start, batch):
        rq = get_advanced_query(self.query, self.last_run, self.current_run)
        return self.slave.getIssues(self.project_id, rq, start, batch)

    def _get_updated_in_master_from_last_run(self, start, batch):
        rq = get_advanced_query(self.query, self.last_run, self.current_run)
        return self.master.getIssues(self.project_id, rq, start, batch)

    def _mark_issues_as_sync(self, master_issue_number, master_issue_id, slave_issue_id):
        self.master_executor.executeCommand(master_issue_id, "tag " + tag)
        self.slave_executor.executeCommand(slave_issue_id, "tag " + tag)
        self.slave_executor.executeCommand(slave_issue_id, master_sync_field_name + " " + master_issue_number)
        self.issue_binder.addBinding(master_issue_id, slave_issue_id)

    def _create_and_attach_sync_field(self, yt, project_id, sync_field_name):
        sync_field_created = any(field.name == sync_field_name for field in yt.getCustomFields())
        if not sync_field_created:
            yt.createCustomFieldDetailed(sync_field_name, "integer", False, True)
        sync_field_attached = any(field.name == sync_field_name for field in yt.getProjectCustomFields(project_id))
        if not sync_field_attached:
            yt.createProjectCustomFieldDetailed(project_id, sync_field_name, empty_field_text)



import calendar
import functools
import httplib2
import json
import re
import sys
import tempfile
import time
import urllib.request, urllib.parse, urllib.error
from xml.dom import Node
from xml.dom import minidom
from xml.sax.saxutils import escape, quoteattr
import datetime
import youtrack

def relogin_on_401(f):
    @functools.wraps(f)
    def wrapped(self, *args, **kwargs):
        attempts = 10
        while attempts:
            try:
                return f(self, *args, **kwargs)
            except youtrack.YouTrackException as e:
                if e.response.status not in (401, 403, 500, 504):
                    raise e
                if e.response.status == 504:
                    time.sleep(30)
                elif self._last_credentials is not None:
                    self._login(*self._last_credentials)
                else:
                    break
                attempts -= 1
        return f(self, *args, **kwargs)
    return wrapped


class Connection(object):
    def __init__(self, url, login=None, password=None, proxy_info=None, token=None):
        if proxy_info is None:
            self.http = httplib2.Http(disable_ssl_certificate_validation=True)
        else:
            self.http = httplib2.Http(disable_ssl_certificate_validation=True,
                                      proxy_info=proxy_info)

        self.url = url.rstrip('/')
        self.baseUrl = self.url + "/rest"
        self.headers = dict()
        self._last_credentials = None

        if token:
            self.set_auth_token(token)
        elif login:
            self._login(login, password)

    def set_auth_token(self, token):
        if token:
            self.headers = {'Authorization': 'Bearer ' + token}

    def _login(self, login, password):
        if login is None:
            login = ''
        if password is None:
            password = ''
        body = 'login=%s&password=%s' % (urllib.parse.quote(login), urllib.parse.quote(password))
        response, content = self.http.request(
            uri=self.baseUrl + '/user/login',
            method='POST',
            body=body,
            headers={'Connection': 'keep-alive',
                     'Content-Type': 'application/x-www-form-urlencoded',
                     'Content-Length': str(len(body))}
        )
        if response.status != 200:
            raise youtrack.YouTrackException('/user/login', response, content)
        self.headers = {'Cookie': response['set-cookie'],
                        'Cache-Control': 'no-cache'}
        self._last_credentials = (login, password)

    @staticmethod
    def __get_illegal_xml_chars_re():
        _illegal_unichrs = [(0x00, 0x08), (0x0B, 0x0C), (0x0E, 0x1F),
                            (0x7F, 0x84), (0x86, 0x9F), (0xFDD0, 0xFDDF),
                            (0xFFFE, 0xFFFF)]
        _illegal_ranges = ["%s-%s" % (chr(low), chr(high))
                           for (low, high) in _illegal_unichrs]
        return re.compile(b'[%s]' % ''.join(_illegal_ranges))

    @relogin_on_401
    def _req(self, method, url, body=None, ignoreStatus=None, content_type=None):
        headers = self.headers
        if method == 'PUT' or method == 'POST':
            headers = headers.copy()
            if body:
                if content_type is None:
                    content_type = 'application/xml; charset=UTF-8'

                #if content_type.lower().find('/xml') != -1:
                    # Remove invalid xml/utf-8 data
                #    body = re.sub(self.__get_illegal_xml_chars_re(), b'', body)

                headers['Content-Type'] = content_type
                headers['Content-Length'] = len(body)
        elif method == 'GET' and content_type is not None:
            headers = headers.copy()
            headers['Accept'] = content_type

        if url.startswith('http'):
            response, content = self.http.request(
                url,
                method,
                headers=headers,
                body=body)
        else:
            response, content = self.http.request(
                (self.baseUrl + url),
                method,
                headers=headers,
                body=body)

        #if response.get('content-type', '').lower().find('/xml') != -1:
        #    # Remove invalid xml/utf-8 data
        #    content = re.sub(self.__get_illegal_xml_chars_re(), b'', content)

        # TODO: Why do we need this?
        #content = content.translate(None, '\0')
        content = re.sub(b'system_user[%@][a-zA-Z0-9]+', b'guest', content)

        if response.status not in (200, 201) and \
                (ignoreStatus != response.status):
            raise youtrack.YouTrackException(url, response, content)

        return response, content

    def _reqXml(self, method, url, body=None, ignoreStatus=None):
        response, content = self._req(
            method, url, body, ignoreStatus, "application/xml")
        if "content-type" in response:
            if response["content-type"].find("/xml") != -1 and content:
                try:
                    return minidom.parseString(content)
                except Exception as e:
                    print((str(e)))
                    return ""
            elif response["content-type"].find("/json") != -1 and content:
                try:
                    return json.loads(content)
                except Exception as e:
                    print((str(e)))
                    return ""

        if method == 'PUT' and ('location' in response):
            return 'Created: ' + response['location']
        else:
            return content

    def _get(self, url):
        return self._reqXml('GET', url)

    def _put(self, url):
        return self._reqXml('PUT', url, '<empty/>\n\n')

    def getIssue(self, id):
        return youtrack.Issue(self._get("/issue/" + id), self)

    def createIssue(self, project, assignee, summary, description, priority=None, type=None, subsystem=None, state=None,
                    affectsVersion=None,
                    fixedVersion=None, fixedInBuild=None, permittedGroup=None):
        params = {'project': project,
                  'summary': summary}
        if description is not None:
            params['description'] = description
        if assignee is not None:
            params['assignee'] = assignee
        if priority is not None:
            params['priority'] = priority
        if type is not None:
            params['type'] = type
        if subsystem is not None:
            params['subsystem'] = subsystem
        if state is not None:
            params['state'] = state
        if affectsVersion is not None:
            params['affectsVersion'] = affectsVersion
        if fixedVersion is not None:
            params['fixVersion'] = fixedVersion
        if fixedInBuild is not None:
            params['fixedInBuild'] = fixedInBuild
        if permittedGroup is not None:
            params['permittedGroup'] = permittedGroup

        return self._req('PUT', '/issue', urllib.parse.urlencode(params), content_type='application/x-www-form-urlencoded')

    def deleteIssue(self, issue_id):
        return self._req('DELETE', '/issue/%s' % issue_id)

    def get_changes_for_issue(self, issue):
        return [youtrack.IssueChange(change, self) for change in
                self._get("/issue/%s/changes" % issue).getElementsByTagName('change')]

    def getComments(self, id):
        response, content = self._req('GET', '/issue/' + id + '/comment')
        xml = minidom.parseString(content)
        return [youtrack.Comment(e, self) for e in xml.documentElement.childNodes if e.nodeType == Node.ELEMENT_NODE]

    def getAttachments(self, id):
        response, content = self._req('GET', '/issue/' + id + '/attachment')
        xml = minidom.parseString(content)
        return [youtrack.Attachment(e, self) for e in xml.documentElement.childNodes if e.nodeType == Node.ELEMENT_NODE]

    def getAttachmentContent(self, url):
        f = urllib.request.urlopen(urllib.request.Request(self.url + url, headers=self.headers))
        return f

    def deleteAttachment(self, issue_id, attachment_id):
        return self._req('DELETE', '/issue/%s/attachment/%s' % (issue_id, attachment_id))

    def createAttachmentFromAttachment(self, issueId, a):
        try:
            content = a.getContent()
            contentLength = None
            if 'content-length' in content.headers.dict:
                contentLength = int(content.headers.dict['content-length'])
            print('Importing attachment for issue ', issueId)
            try:
                print('Name: ', a.name)
            except Exception as e:
                print(e)
            try:
                print('Author: ', a.authorLogin)
            except Exception as e:
                print(e)
            return self.importAttachment(issueId, a.name, content, a.authorLogin,
                contentLength=contentLength,
                contentType=content.info().type,
                created=a.created if hasattr(a, 'created') else None,
                group=a.group if hasattr(a, 'group') else '')
        except urllib.error.HTTPError as e:
            print("Can't create attachment")
            try:
                err_content = e.read()
                issue_id = issueId
                attach_name = a.name
                attach_url = a.url
                if isinstance(err_content, str):
                    err_content = err_content.encode('utf-8')
                if isinstance(issue_id, str):
                    issue_id = issue_id.encode('utf-8')
                if isinstance(attach_name, str):
                    attach_name = attach_name.encode('utf-8')
                if isinstance(attach_url, str):
                    attach_url = attach_url.encode('utf-8')
                print("HTTP CODE: ", e.code)
                print("REASON: ", err_content)
                print("IssueId: ", issue_id)
                print("Attachment filename: ", attach_name)
                print("Attachment URL: ", attach_url)
            except Exception:
                pass
        except Exception as e:
            try:
                print(content.geturl())
                print(content.getcode())
                print(content.info())
            except Exception:
                pass
            raise e
            

    def _process_attachments(self, authorLogin, content, contentLength, contentType, created, group, issueId, name,
                             url_prefix='/issue/'):
        if contentType is not None:
            content.contentType = contentType
        if contentLength is not None:
            content.contentLength = contentLength
        elif not isinstance(content, file):
            tmp = tempfile.NamedTemporaryFile(mode='w+b')
            tmp.write(content.read())
            tmp.flush()
            tmp.seek(0)
            content = tmp

        #post_data = {'attachment': content}
        post_data = {name: content}
        headers = self.headers.copy()
        #headers['Content-Type'] = contentType
        # name without extension to workaround: http://youtrack.jetbrains.net/issue/JT-6110
        params = {#'name': os.path.splitext(name)[0],
                  'authorLogin': authorLogin.encode('utf-8'),
        }
        if group is not None:
            params["group"] = group
        if created is not None:
            params['created'] = created
        else:
            try:
                params['created'] = self.getIssue(issueId).created
            except youtrack.YouTrackException:
                params['created'] = str(calendar.timegm(datetime.now().timetuple()) * 1000)

        url = self.baseUrl + url_prefix + issueId + "/attachment?" + urllib.parse.urlencode(params)
        r = urllib.request.Request(url,
            headers=headers, data=post_data)
        #r.set_proxy('localhost:8888', 'http')
        try:
            res = urllib.request.urlopen(r)
        except urllib.error.HTTPError as e:
            if e.code == 201:
                return e.msg + ' ' + name
            raise e
        return res

    def createAttachment(self, issueId, name, content, authorLogin='', contentType=None, contentLength=None,
                         created=None, group=''):
        return self._process_attachments(authorLogin, content, contentLength, contentType, created, group, issueId,
                                         name)

    def importAttachment(self, issue_id, name, content, authorLogin, contentType, contentLength, created=None,
                         group=''):
        return self._process_attachments(authorLogin, content, contentLength, contentType, created, group, issue_id,
                                         name, '/import/')


    def getLinks(self, id, outwardOnly=False):
        response, content = self._req('GET', '/issue/' + urllib.parse.quote(id) + '/link')
        xml = minidom.parseString(content)
        res = []
        for c in [e for e in xml.documentElement.childNodes if e.nodeType == Node.ELEMENT_NODE]:
            link = youtrack.Link(c, self)
            if link.source == id or not outwardOnly:
                res.append(link)
        return res

    def getUser(self, login):
        """ http://confluence.jetbrains.net/display/YTD2/GET+user
        """
        if login.startswith('system_user'):
            login = 'guest'
        return youtrack.User(self._get("/admin/user/" + urllib.parse.quote(login.encode('utf8'))), self)

    def createUser(self, user):
        """ user from getUser
        """
        # self.createUserDetailed(user.login, user.fullName, user.email, user.jabber)
        self.importUsers([user])

    def createUserDetailed(self, login, fullName, email, jabber):
        self.importUsers([{'login': login, 'fullName': fullName, 'email': email, 'jabber': jabber}])

    #        return self._put('/admin/user/' + login + '?' +
    #                         'password=' + password +
    #                         '&fullName=' + fullName +
    #                         '&email=' + email +
    #                         '&jabber=' + jabber)


    def importUsers(self, users):
        """ Import users, returns import result (http://confluence.jetbrains.net/display/YTD2/Import+Users)
            Example: importUsers([{'login':'vadim', 'fullName':'vadim', 'email':'eee@ss.com', 'jabber':'fff@fff.com'},
                                  {'login':'maxim', 'fullName':'maxim', 'email':'aaa@ss.com', 'jabber':'www@fff.com'}])
        """
        if len(users) <= 0: return

        known_attrs = ('login', 'fullName', 'email', 'jabber')

        xml = '<list>\n'
        for u in users:
            xml += '  <user ' + "".join(k + '=' + quoteattr(u[k]) + ' ' for k in u if k in known_attrs) + '/>\n'
        xml += '</list>'
        #TODO: convert response xml into python objects
        if isinstance(xml, str):
            xml = xml.encode('utf-8')
        return self._reqXml('PUT', '/import/users', xml, 400).toxml()

    def importIssuesXml(self, projectId, assigneeGroup, xml):
        return self._reqXml('PUT', '/import/' + urllib.parse.quote(projectId) + '/issues?' +
                                   urllib.parse.urlencode({'assigneeGroup': assigneeGroup}),
            xml, 400).toxml()

    def importLinks(self, links):
        """ Import links, returns import result (http://confluence.jetbrains.net/display/YTD2/Import+Links)
            Accepts result of getLinks()
            Example: importLinks([{'login':'vadim', 'fullName':'vadim', 'email':'eee@ss.com', 'jabber':'fff@fff.com'},
                                  {'login':'maxim', 'fullName':'maxim', 'email':'aaa@ss.com', 'jabber':'www@fff.com'}])
        """
        xml = '<list>\n'
        for l in links:
            # ignore typeOutward and typeInward returned by getLinks()
            xml += '  <link ' + "".join(attr + '=' + quoteattr(l[attr]) +
                                        ' ' for attr in l if attr not in ['typeInward', 'typeOutward']) + '/>\n'
        xml += '</list>'
        #TODO: convert response xml into python objects
        res = self._reqXml('PUT', '/import/links', xml, 400)
        return res.toxml() if hasattr(res, "toxml") else res

    def importIssues(self, projectId, assigneeGroup, issues):
        """ Import issues, returns import result (http://confluence.jetbrains.net/display/YTD2/Import+Issues)
            Accepts retrun of getIssues()
            Example: importIssues([{'numberInProject':'1', 'summary':'some problem', 'description':'some description', 'priority':'1',
                                    'fixedVersion':['1.0', '2.0'],
                                    'comment':[{'author':'yamaxim', 'text':'comment text', 'created':'1267030230127'}]},
                                   {'numberInProject':'2', 'summary':'some problem', 'description':'some description', 'priority':'1'}])
        """
        if len(issues) <= 0:
            return

        bad_fields = ['id', 'projectShortName', 'votes', 'commentsCount',
                      'historyUpdated', 'updatedByFullName', 'updaterFullName',
                      'reporterFullName', 'links', 'attachments', 'jiraId',
                      'entityId', 'tags', 'sprint', 'wikified']

        tt_settings = self.getProjectTimeTrackingSettings(projectId)
        if tt_settings and tt_settings.Enabled and tt_settings.TimeSpentField:
            bad_fields.append(tt_settings.TimeSpentField)

        if not self.isMarkdownSupported():
            bad_fields.append('markdown')

        xml = '<issues>\n'
        issue_records = dict([])

        for issue in issues:
            record = ""
            record += '  <issue>\n'

            comments = None
            if getattr(issue, "getComments", None):
                comments = issue.getComments()

            for issueAttr in issue:
                attrValue = issue[issueAttr]
                if attrValue is None:
                    continue
                if isinstance(attrValue, str):
                    attrValue = attrValue.encode('utf-8')
                if isinstance(issueAttr, str):
                    issueAttr = issueAttr.encode('utf-8')
                if issueAttr == 'comments':
                    comments = attrValue
                else:
                    # ignore bad fields from getIssue()
                    if issueAttr not in bad_fields:
                        record += '    <field name="' + issueAttr + '">\n'
                        if isinstance(attrValue, list) or getattr(attrValue, '__iter__', False):
                            for v in attrValue:
                                if isinstance(v, str):
                                    v = v.encode('utf-8')
                                record += '      <value>' + escape(v.strip()) + '</value>\n'
                        else:
                            record += '      <value>' + escape(attrValue.strip()) + '</value>\n'
                        record += '    </field>\n'

            if comments:
                for comment in comments:
                    record += '    <comment'
                    for ca in comment:
                        val = comment[ca]
                        if isinstance(ca, str):
                            ca = ca.encode('utf-8')
                        if isinstance(val, str):
                            val = val.encode('utf-8')
                        record += ' ' + ca + '=' + quoteattr(val, {"\n" : "&#xA;"})
                    record += '/>\n'

            record += '  </issue>\n'
            xml += record
            issue_records[issue.numberInProject] = record

        xml += '</issues>'

        #print xml
        #TODO: convert response xml into python objects

        if isinstance(xml, str):
            xml = xml.encode('utf-8')

        if isinstance(assigneeGroup, str):
            assigneeGroup = assigneeGroup.encode('utf-8')

        url = '/import/' + urllib.parse.quote(projectId) + '/issues?' + urllib.parse.urlencode({'assigneeGroup': assigneeGroup})
        if isinstance(url, str):
            url = url.encode('utf-8')
        result = self._reqXml('PUT', url, xml, 400)
        if (result == "") and (len(issues) > 1):
            for issue in issues:
                self.importIssues(projectId, assigneeGroup, [issue])
        response = ""
        try:
            response = result.toxml().encode('utf-8')
        except:
            sys.stderr.write("can't parse response")
            sys.stderr.write("request was")
            sys.stderr.write(xml)
            return response
        item_elements = minidom.parseString(response).getElementsByTagName("item")
        if len(item_elements) != len(issues):
            sys.stderr.write(response)
        else:
            for item in item_elements:
                id = item.attributes["id"].value
                imported = item.attributes["imported"].value.lower()
                if imported == "true":
                    print("Issue [ %s-%s ] imported successfully" % (projectId, id))
                else:
                    sys.stderr.write("")
                    sys.stderr.write("Failed to import issue [ %s-%s ]." % (projectId, id))
                    sys.stderr.write("Reason : ")
                    sys.stderr.write(item.toxml())
                    sys.stderr.write("Request was :")
                    if isinstance(issue_records[id], str):
                        sys.stderr.write(issue_records[id].encode('utf-8'))
                    else:
                        sys.stderr.write(issue_records[id])
                print("")
        return response

    def getProjects(self):
        projects = {}
        for e in self._get("/project/all").documentElement.childNodes:
            projects[e.getAttribute('shortName')] = e.getAttribute('name')
        return projects

    def getProject(self, projectId):
        """ http://confluence.jetbrains.net/display/YTD2/GET+project
        """
        return youtrack.Project(self._get("/admin/project/" + urllib.parse.quote(projectId)), self)

    def getProjectIds(self):
        response, content = self._req('GET', '/admin/project/')
        xml = minidom.parseString(content)
        return [e.getAttribute('id') for e in xml.documentElement.childNodes if e.nodeType == Node.ELEMENT_NODE]

    def getProjectAssigneeGroups(self, projectId):
        response, content = self._req('GET', '/admin/project/' + urllib.parse.quote(projectId) + '/assignee/group')
        xml = minidom.parseString(content)
        return [youtrack.Group(e, self) for e in xml.documentElement.childNodes if e.nodeType == Node.ELEMENT_NODE]

    def getGroup(self, name):
        return youtrack.Group(self._get("/admin/group/" + urllib.parse.quote(name.encode('utf-8'))), self)

    def getGroups(self):
        response, content = self._req('GET', '/admin/group')
        xml = minidom.parseString(content)
        return [youtrack.Group(e, self) for e in xml.documentElement.childNodes if e.nodeType == Node.ELEMENT_NODE]

    def deleteGroup(self, name):
        return self._req('DELETE', "/admin/group/" + urllib.parse.quote(name.encode('utf-8')))

    def getUserGroups(self, userName):
        response, content = self._req('GET', '/admin/user/%s/group' % urllib.parse.quote(userName.encode('utf-8')))
        xml = minidom.parseString(content)
        return [youtrack.Group(e, self) for e in xml.documentElement.childNodes if e.nodeType == Node.ELEMENT_NODE]

    def setUserGroup(self, user_name, group_name):
        if isinstance(user_name, str):
            user_name = user_name.encode('utf-8')
        if isinstance(group_name, str):
            group_name = group_name.encode('utf-8')
        response, content = self._req('POST',
            '/admin/user/%s/group/%s' % (urllib.parse.quote(user_name), urllib.parse.quote(group_name)),
            body='')
        return response

    def createGroup(self, group):
        content = self._put(
            '/admin/group/%s?autoJoin=false' % urllib.parse.quote(group.name))
        return content

    def addUserRoleToGroup(self, group, userRole):
        url_group_name = urllib.parse.quote(group.name)
        url_role_name = urllib.parse.quote(userRole.name)
        response, content = self._req('PUT', '/admin/group/%s/role/%s' % (url_group_name, url_role_name),
            body=userRole.toXml())
        return content

    def getRole(self, name):
        return youtrack.Role(self._get("/admin/role/" + urllib.parse.quote(name)), self)

    def getRoles(self):
        response, content = self._req('GET', '/admin/role')
        xml = minidom.parseString(content)
        return [youtrack.Role(e, self) for e in xml.documentElement.childNodes if e.nodeType == Node.ELEMENT_NODE]

    def getGroupRoles(self, group_name):
        response, content = self._req('GET', '/admin/group/%s/role' % urllib.parse.quote(group_name))
        xml = minidom.parseString(content)
        return [youtrack.UserRole(e, self) for e in xml.documentElement.childNodes if e.nodeType == Node.ELEMENT_NODE]

    def createRole(self, role):
        url_role_name = urllib.parse.quote(role.name)
        url_role_dscr = ''
        if hasattr(role, 'description'):
                url_role_dscr = urllib.parse.quote(role.description)
        content = self._put('/admin/role/%s?description=%s' % (url_role_name, url_role_dscr))
        return content

    def changeRole(self, role, new_name, new_description):
        url_role_name = urllib.parse.quote(role.name)
        url_new_name = urllib.parse.quote(new_name)
        url_new_dscr = urllib.parse.quote(new_description)
        content = self._req('POST',
            '/admin/role/%s?newName=%s&description=%s' % (url_role_name, url_new_name, url_new_dscr))
        return content

    def addPermissionToRole(self, role, permission):
        url_role_name = urllib.parse.quote(role.name)
        url_prm_name = urllib.parse.quote(permission.name)
        content = self._req('POST', '/admin/role/%s/permission/%s' % (url_role_name, url_prm_name))
        return content

    def getRolePermissions(self, role):
        response, content = self._req('GET', '/admin/role/%s/permission' % urllib.parse.quote(role.name))
        xml = minidom.parseString(content)
        return [youtrack.Permission(e, self) for e in xml.documentElement.childNodes if e.nodeType == Node.ELEMENT_NODE]

    def getPermissions(self):
        response, content = self._req('GET', '/admin/permission')
        xml = minidom.parseString(content)
        return [youtrack.Permission(e, self) for e in xml.documentElement.childNodes if e.nodeType == Node.ELEMENT_NODE]

    def getSubsystem(self, projectId, name):
        response, content = self._req('GET', '/admin/project/' + projectId + '/subsystem/' + urllib.parse.quote(name))
        xml = minidom.parseString(content)
        return youtrack.Subsystem(xml, self)

    def getSubsystems(self, projectId):
        response, content = self._req('GET', '/admin/project/' + projectId + '/subsystem')
        xml = minidom.parseString(content)
        return [youtrack.Subsystem(e, self) for e in xml.documentElement.childNodes if e.nodeType == Node.ELEMENT_NODE]

    def getVersions(self, projectId):
        response, content = self._req('GET', '/admin/project/' + urllib.parse.quote(projectId) + '/version?showReleased=true')
        xml = minidom.parseString(content)
        return [self.getVersion(projectId, v.getAttribute('name')) for v in
                xml.documentElement.getElementsByTagName('version')]

    def getVersion(self, projectId, name):
        return youtrack.Version(
            self._get("/admin/project/" + urllib.parse.quote(projectId) + "/version/" + urllib.parse.quote(name)), self)

    def getBuilds(self, projectId):
        response, content = self._req('GET', '/admin/project/' + urllib.parse.quote(projectId) + '/build')
        xml = minidom.parseString(content)
        return [youtrack.Build(e, self) for e in xml.documentElement.childNodes if e.nodeType == Node.ELEMENT_NODE]


    def getUsers(self, params={}):
        first = True
        users = []
        position = 0
        user_search_params = urllib.parse.urlencode(params)
        while True:
            response, content = self._req('GET', "/admin/user/?start=%s&%s" % (str(position), user_search_params))
            position += 10
            xml = minidom.parseString(content)
            newUsers = [youtrack.User(e, self) for e in xml.documentElement.childNodes if
                        e.nodeType == Node.ELEMENT_NODE]
            if not len(newUsers): return users
            users += newUsers


    def getUsersTen(self, start):
        response, content = self._req('GET', "/admin/user/?start=%s" % str(start))
        xml = minidom.parseString(content)
        users = [youtrack.User(e, self) for e in xml.documentElement.childNodes if
                 e.nodeType == Node.ELEMENT_NODE]
        return users

    def deleteUser(self, login):
        return self._req('DELETE', "/admin/user/" + urllib.parse.quote(login.encode('utf-8')))

    # TODO this function is deprecated
    def createBuild(self):
        raise NotImplementedError

    # TODO this function is deprecated
    def createBuilds(self):
        raise NotImplementedError

    def createProject(self, project):
        return self.createProjectDetailed(project.id, project.name, project.description, project.lead)

    def deleteProject(self, projectId):
        return self._req('DELETE', "/admin/project/" + urllib.parse.quote(projectId))

    def createProjectDetailed(self, projectId, name, description, projectLeadLogin, startingNumber=1):
        _name = name
        _desc = description
        if isinstance(_name, str):
            _name = _name.encode('utf-8')
        if isinstance(_desc, str):
            _desc = _desc.encode('utf-8')
        _name = _name.replace('/', ' ')
        return self._put('/admin/project/' + projectId + '?' +
                         urllib.parse.urlencode({'projectName': _name,
                                           'description': _desc + ' ',
                                           'projectLeadLogin': projectLeadLogin,
                                           'lead': projectLeadLogin,
                                           'startingNumber': str(startingNumber)}))

    # TODO this function is deprecated
    def createSubsystems(self, projectId, subsystems):
        """ Accepts result of getSubsystems()
        """

        for s in subsystems:
            self.createSubsystem(projectId, s)

    # TODO this function is deprecated
    def createSubsystem(self, projectId, s):
        return self.createSubsystemDetailed(projectId, s.name, s.isDefault,
            s.defaultAssignee if s.defaultAssignee != '<no user>' else '')

    # TODO this function is deprecated
    def createSubsystemDetailed(self, projectId, name, isDefault, defaultAssigneeLogin):
        self._put('/admin/project/' + projectId + '/subsystem/' + urllib.parse.quote(name.encode('utf-8')) + "?" +
                  urllib.parse.urlencode({'isDefault': str(isDefault),
                                    'defaultAssignee': defaultAssigneeLogin}))

        return 'Created'

    # TODO this function is deprecated
    def deleteSubsystem(self, projectId, name):
        return self._reqXml('DELETE', '/admin/project/' + projectId + '/subsystem/' + urllib.parse.quote(name.encode('utf-8'))
            , '')

    # TODO this function is deprecated
    def createVersions(self, projectId, versions):
        """ Accepts result of getVersions()
        """

        for v in versions:
            self.createVersion(projectId, v)

    # TODO this function is deprecated
    def createVersion(self, projectId, v):
        return self.createVersionDetailed(projectId, v.name, v.isReleased, v.isArchived, releaseDate=v.releaseDate,
            description=v.description)

    # TODO this function is deprecated
    def createVersionDetailed(self, projectId, name, isReleased, isArchived, releaseDate=None, description=''):
        params = {'description': description,
                  'isReleased': str(isReleased),
                  'isArchived': str(isArchived)}
        if releaseDate is not None:
            params['releaseDate'] = str(releaseDate)
        return self._put(
            '/admin/project/' + urllib.parse.quote(projectId) + '/version/' + urllib.parse.quote(name.encode('utf-8')) + "?" +
            urllib.parse.urlencode(params))

    def getIssues(self, projectId, filter, after, max):
        #response, content = self._req('GET', '/project/issues/' + urllib.parse.quote(projectId) + "?" +
        path = '/issue'
        if projectId:
            path += '/byproject/' + urllib.parse.quote(projectId)
        response, content = self._req('GET', path + "?" +
                                             urllib.parse.urlencode({'after': str(after),
                                                               'max': str(max),
                                                               'filter': filter}))
        xml = minidom.parseString(content)
        return [youtrack.Issue(e, self) for e in xml.documentElement.childNodes if e.nodeType == Node.ELEMENT_NODE]

    def getNumberOfIssues(self, filter = '', waitForServer=True):
        while True:
          urlFilterList = [('filter',filter)]
          finalUrl = '/issue/count?' + urllib.parse.urlencode(urlFilterList)
          response, content = self._req('GET', finalUrl, content_type="application/json")
          result = json.loads(content)
          numberOfIssues = result['value']
          if (not waitForServer):
            return numberOfIssues
          if (numberOfIssues!=-1):
            break

        time.sleep(5)
        return self.getNumberOfIssues(filter,False)


    def getAllSprints(self,agileID):
        response, content = self._req('GET', '/agile/' + agileID + "/sprints?")
        xml = minidom.parseString(content)
        return [(e.getAttribute('name'),e.getAttribute('start'),e.getAttribute('finish')) for e in xml.documentElement.childNodes if e.nodeType == Node.ELEMENT_NODE]

    def getAllIssues(self, filter = '', after = 0, max = 999999, withFields = ()):
        urlJobby = [('with',field) for field in withFields] + \
                    [('after',str(after)),
                    ('max',str(max)),
                    ('filter',filter)]
        response, content = self._req('GET', '/issue' + "?" +
                                             urllib.parse.urlencode(urlJobby))
        xml = minidom.parseString(content)
        return [youtrack.Issue(e, self) for e in xml.documentElement.childNodes if e.nodeType == Node.ELEMENT_NODE]

    def exportIssueLinks(self):
        response, content = self._req('GET', '/export/links')
        xml = minidom.parseString(content)
        return [youtrack.Link(e, self) for e in xml.documentElement.childNodes if e.nodeType == Node.ELEMENT_NODE]

    def executeCommand(self, issueId, command, comment=None, group=None, run_as=None, disable_notifications=False):
        if isinstance(command, str):
            command = command.encode('utf-8')
        params = {'command': command}

        if comment is not None:
            params['comment'] = comment

        if group is not None:
            params['group'] = group

        if run_as is not None:
            params['runAs'] = run_as

        if disable_notifications:
            params['disableNotifications'] = disable_notifications

        for p in params:
            if isinstance(params[p], str):
                params[p] = params[p].encode('utf-8')

        self._req('POST',
                  '/issue/' + issueId + "/execute",
                  body=urllib.parse.urlencode(params),
                  content_type='application/x-www-form-urlencoded')

        return "Command executed"

    def getCustomField(self, name):
        return youtrack.CustomField(self._get("/admin/customfield/field/" + urllib.parse.quote(name.encode('utf-8'))), self)

    def getCustomFields(self):
        response, content = self._req('GET', '/admin/customfield/field')
        xml = minidom.parseString(content)
        return [self.getCustomField(e.getAttribute('name')) for e in xml.documentElement.childNodes if
                e.nodeType == Node.ELEMENT_NODE]

    def createCustomField(self, cf):
        params = dict([])
        if hasattr(cf, "defaultBundle"):
            params["defaultBundle"] = cf.defaultBundle
        if hasattr(cf, "attachBundlePolicy"):
            params["attachBundlePolicy"] = cf.attachBundlePolicy
        auto_attached = False
        if hasattr(cf, "autoAttached"):
            auto_attached = cf.autoAttached
        return self.createCustomFieldDetailed(cf.name, cf.type, cf.isPrivate, cf.visibleByDefault, auto_attached,
            params)

    def createCustomFieldDetailed(self, customFieldName, typeName, isPrivate, defaultVisibility,
                                  auto_attached=False, additional_params=dict([])):
        params = {'type': typeName, 'isPrivate': str(isPrivate), 'defaultVisibility': str(defaultVisibility),
                  'autoAttached': str(auto_attached)}
        params.update(additional_params)
        for key in params:
            if isinstance(params[key], str):
                params[key] = params[key].encode('utf-8')

        self._put('/admin/customfield/field/' + urllib.parse.quote(customFieldName.encode('utf-8')) + '?' +
                  urllib.parse.urlencode(params), )

        return "Created"

    def createCustomFields(self, cfs):
        for cf in cfs:
            self.createCustomField(cf)

    def getProjectCustomField(self, projectId, name):
        if isinstance(name, str):
            name = name.encode('utf8')
        return youtrack.ProjectCustomField(
            self._get("/admin/project/" + urllib.parse.quote(projectId) + "/customfield/" + urllib.parse.quote(name))
            , self)

    def getProjectCustomFields(self, projectId):
        response, content = self._req('GET', '/admin/project/' + urllib.parse.quote(projectId) + '/customfield')
        xml = minidom.parseString(content)
        return [self.getProjectCustomField(projectId, e.getAttribute('name')) for e in
                xml.getElementsByTagName('projectCustomField')]

    def createProjectCustomField(self, projectId, pcf):
        return self.createProjectCustomFieldDetailed(projectId, pcf.name, pcf.emptyText, pcf.params)

    def createProjectCustomFieldDetailed(self, projectId, customFieldName, emptyFieldText, params=None):
        if not len(emptyFieldText.strip()):
            emptyFieldText = "No " + customFieldName
        if isinstance(customFieldName, str):
            customFieldName = customFieldName.encode('utf-8')
        _params = {'emptyFieldText': emptyFieldText}
        if params is not None:
            _params.update(params)
        for key in _params:
            if isinstance(_params[key], str):
                _params[key] = _params[key].encode('utf-8')
        return self._put(
            '/admin/project/' + projectId + '/customfield/' + urllib.parse.quote(customFieldName) + '?' +
            urllib.parse.urlencode(_params))

    def deleteProjectCustomField(self, project_id, pcf_name):
        self._req('DELETE', '/admin/project/' + urllib.parse.quote(project_id) + "/customfield/" + urllib.parse.quote(pcf_name))

    def getIssueLinkTypes(self):
        response, content = self._req('GET', '/admin/issueLinkType')
        xml = minidom.parseString(content)
        return [youtrack.IssueLinkType(e, self) for e in xml.getElementsByTagName('issueLinkType')]

    def createIssueLinkTypes(self, issueLinkTypes):
        for ilt in issueLinkTypes:
            return self.createIssueLinkType(ilt)

    def createIssueLinkType(self, ilt):
        return self.createIssueLinkTypeDetailed(ilt.name, ilt.outwardName, ilt.inwardName, ilt.directed)

    def createIssueLinkTypeDetailed(self, name, outwardName, inwardName, directed):
        if isinstance(name, str):
            name = name.encode('utf-8')
        if isinstance(outwardName, str):
            outwardName = outwardName.encode('utf-8')
        if isinstance(inwardName, str):
            inwardName = inwardName.encode('utf-8')
        return self._put('/admin/issueLinkType/' + urllib.parse.quote(name) + '?' +
                         urllib.parse.urlencode({'outwardName': outwardName,
                                           'inwardName': inwardName,
                                           'directed': directed}))

    def getEvents(self, issue_id):
        return self._get('/event/issueEvents/' + urllib.parse.quote(issue_id))

    def getWorkItems(self, issue_id):
        try:
            response, content = self._req('GET',
                '/issue/%s/timetracking/workitem' % urllib.parse.quote(issue_id), content_type="application/xml")
            xml = minidom.parseString(content)
            return [youtrack.WorkItem(e, self) for e in xml.documentElement.childNodes if
                    e.nodeType == Node.ELEMENT_NODE]
        except youtrack.YouTrackException as e:
            print("Can't get work items.", str(e))
            return []


    def createWorkItem(self, issue_id, work_item):
        xml =  '<workItem>'
        xml += '<date>%s</date>' % work_item.date
        xml += '<duration>%s</duration>' % work_item.duration
        if hasattr(work_item, 'description') and work_item.description is not None:
            xml += '<description>%s</description>' % escape(work_item.description)
        if hasattr(work_item, 'worktype') and work_item.worktype is not None:
            xml += '<worktype><name>%s</name></worktype>' % work_item.worktype
        xml += '</workItem>'
        if isinstance(xml, str):
            xml = xml.encode('utf-8')
        self._reqXml('POST',
            '/issue/%s/timetracking/workitem' % urllib.parse.quote(issue_id), xml)

    def importWorkItems(self, issue_id, work_items):
        xml = ''
        for work_item in work_items:
            xml +=  '<workItem>'
            xml += '<date>%s</date>' % work_item.date
            xml += '<duration>%s</duration>' % work_item.duration
            if hasattr(work_item, 'description') and work_item.description is not None:
                xml += '<description>%s</description>' % escape(work_item.description)
            if hasattr(work_item, 'worktype') and work_item.worktype is not None:
                xml += '<worktype><name>%s</name></worktype>' % work_item.worktype
            xml += '<author login=%s></author>' % quoteattr(work_item.authorLogin)
            xml += '</workItem>'
        if isinstance(xml, str):
            xml = xml.encode('utf-8')
        if xml:
            xml = '<workItems>' + xml + '</workItems>'
            try:
                self.headers['Accept'] = 'application/xml'
                self._reqXml(
                    'PUT',
                    '/import/issue/%s/workitems' % urllib.parse.quote(issue_id), xml)
            finally:
                del self.headers['Accept']

    def getSearchIntelliSense(self, query,
                              context=None, caret=None, options_limit=None):
        opts = {'filter': query}
        if context:
            opts['project'] = context
        if caret is not None:
            opts['caret'] = caret
        if options_limit is not None:
            opts['optionsLimit'] = options_limit
        return youtrack.IntelliSense(
            self._get('/issue/intellisense?' + urllib.parse.urlencode(opts)), self)

    def getCommandIntelliSense(self, issue_id, command,
                               run_as=None, caret=None, options_limit=None):
        opts = {'command': command}
        if run_as:
            opts['runAs'] = run_as
        if caret is not None:
            opts['caret'] = caret
        if options_limit is not None:
            opts['optionsLimit'] = options_limit
        return youtrack.IntelliSense(
            self._get('/issue/%s/execute/intellisense?%s'
                      % (issue_id, urllib.parse.urlencode(opts))), self)

    def getGlobalTimeTrackingSettings(self):
        try:
            cont = self._get('/admin/timetracking')
            return youtrack.GlobalTimeTrackingSettings(cont, self)
        except youtrack.YouTrackException as e:
            if e.response.status != 404:
                raise e

    def getProjectTimeTrackingSettings(self, projectId):
        try:
            cont = self._get('/admin/project/' + projectId + '/timetracking')
            return youtrack.ProjectTimeTrackingSettings(cont, self)
        except youtrack.YouTrackException as e:
            if e.response.status != 404:
                raise e

    def setGlobalTimeTrackingSettings(self, daysAWeek=None, hoursADay=None):
        xml = '<timesettings>'
        if daysAWeek is not None:
            xml += '<daysAWeek>%d</daysAWeek>' % daysAWeek
        if hoursADay is not None:
            xml += '<hoursADay>%d</hoursADay>' % hoursADay
        xml += '</timesettings>'
        return self._reqXml('PUT', '/admin/timetracking', xml)

    def setProjectTimeTrackingSettings(self,
        projectId, estimateField=None, timeSpentField=None, enabled=None):
        if enabled is not None:
            xml = '<settings enabled="%s">' % str(enabled == True).lower()
        else:
            xml = '<settings>'
        if estimateField is not None and estimateField != '':
            if isinstance(estimateField, str):
                estimateField = estimateField.encode('utf-8')
            xml += '<estimation name="%s"/>' % estimateField
        if timeSpentField is not None and timeSpentField != '':
            if isinstance(timeSpentField, str):
                timeSpentField = timeSpentField.encode('utf-8')
            xml += '<spentTime name="%s"/>' % timeSpentField
        xml += '</settings>'
        return self._reqXml(
            'PUT', '/admin/project/' + projectId + '/timetracking', xml)

    def get_work_types(self, project_id=None):
        if project_id:
            path = '/admin/project/%s/timetracking/worktype' % project_id
        else:
            path = '/admin/timetracking/worktype'
        try:
            xml = self._get(path)
            return [youtrack.WorkType(e, self)
                    for e in xml.documentElement.childNodes
                    if e.nodeType == Node.ELEMENT_NODE]
        except youtrack.YouTrackException as e:
            print(("Can't get work types", str(e)))
            return []

    def create_work_type(self, name=None, auto_attached=None, work_type=None):
        if work_type:
            wt = work_type
        else:
            if not name:
                raise ValueError("Work type name cannot be empty")
            wt = youtrack.WorkType()
            wt.name = name
            wt.autoAttached = auto_attached
        response, content = self._req(
            'POST', '/admin/timetracking/worktype', wt.toXml())
        return youtrack.WorkType(self._get(response['location']))

    def create_work_type_safe(self,
                              name=None, auto_attached=None, work_type=None):
        try:
            return self.create_work_type(name, auto_attached, work_type)
        except youtrack.YouTrackException as e:
            # Assume that this caused by not unique value and try to find
            # original work type
            if e.response.status not in (400, 409):
                raise e
            if work_type:
                name_lc = work_type.name.lower()
            else:
                name_lc = name.lower()
            for wt in self.get_work_types():
                if wt.name.lower() == name_lc:
                    return wt
            raise e

    def attach_work_type_to_project(self, project_id, work_type_id):
        self._req('PUT',
                  '/admin/project/%s/timetracking/worktype/%s' %
                  (project_id, work_type_id))

    def create_project_work_type(
            self, project_id, name=None, auto_attached=None, work_type=None):
        wt = self.create_work_type_safe(name, auto_attached, work_type)
        self.attach_work_type_to_project(project_id, wt.id)

    def getAllBundles(self, field_type):
        field_type = self.get_field_type(field_type)
        if field_type == "enum":
            tag_name = "enumFieldBundle"
        elif field_type == "user":
            tag_name = "userFieldBundle"
        else:
            tag_name = self.bundle_paths[field_type]
        names = [e.getAttribute("name") for e in self._get('/admin/customfield/' +
                                                           self.bundle_paths[field_type]).getElementsByTagName(
            tag_name)]
        return [self.getBundle(field_type, name) for name in names]


    def get_field_type(self, field_type):
        if "[" in field_type:
            field_type = field_type[0:-3]
        return field_type

    def getBundle(self, field_type, name):
        field_type = self.get_field_type(field_type)
        response = self._get('/admin/customfield/%s/%s' % (self.bundle_paths[field_type],
                                                           urllib.parse.quote(name.encode('utf-8'))))
        return self.bundle_types[field_type](response, self)

    def renameBundle(self, bundle, new_name):
        response, content = self._req("POST", "/admin/customfield/%s/%s?newName=%s" % (
            self.bundle_paths[bundle.get_field_type()], bundle.name, new_name), "", ignoreStatus=301)
        return response

    def createBundle(self, bundle):
        return self._reqXml('PUT', '/admin/customfield/' + self.bundle_paths[bundle.get_field_type()],
            body=bundle.toXml(), ignoreStatus=400)

    def deleteBundle(self, bundle):
        response, content = self._req("DELETE", "/admin/customfield/%s/%s" % (
            self.bundle_paths[bundle.get_field_type()], bundle.name), "")
        return response

    def addValueToBundle(self, bundle, value):
        request = ""
        if bundle.get_field_type() != "user":
            request = "/admin/customfield/%s/%s/" % (
                self.bundle_paths[bundle.get_field_type()], urllib.parse.quote(bundle.name.encode('utf-8')))
            if isinstance(value, str):
                request += urllib.parse.quote(value)
            elif isinstance(value, str):
                request += urllib.parse.quote(value.encode('utf-8'))
            else:
                request += urllib.parse.quote(value.name.encode('utf-8')) + "?"
                params = dict()
                for e in value:
                    if (e != "name") and (e != "element_name") and len(value[e]):
                        if isinstance(value[e], str):
                            params[e] = value[e].encode('utf-8')
                        else:
                            params[e] = value[e]
                if len(params):
                    request += urllib.parse.urlencode(params)
        else:
            request = "/admin/customfield/userBundle/%s/" % urllib.parse.quote(bundle.name.encode('utf-8'))
            if isinstance(value, youtrack.User):
                request += "individual/%s/" % value.login
            elif isinstance(value, youtrack.Group):
                request += "group/%s/" % urllib.parse.quote(value.name.encode('utf-8'))
            else:
                request += "individual/%s/" % urllib.parse.quote(value)
        return self._put(request)

    def removeValueFromBundle(self, bundle, value):
        field_type = bundle.get_field_type()
        request = "/admin/customfield/%s/%s/" % (self.bundle_paths[field_type], bundle.name)
        if field_type != "user":
            request += urllib.parse.quote(value.name)
        elif isinstance(value, youtrack.User):
            request += "individual/" + urllib.parse.quote(value.login)
        else:
            request += "group/" + value.name
        response, content = self._req("DELETE", request, "", ignoreStatus=204)
        return response


    def getEnumBundle(self, name):
        return youtrack.EnumBundle(self._get("/admin/customfield/bundle/" + urllib.parse.quote(name)), self)


    def createEnumBundle(self, eb):
        return self.createBundle(eb)

    def deleteEnumBundle(self, name):
        return self.deleteBundle(self.getEnumBundle(name))

    def createEnumBundleDetailed(self, name, values):
        xml = '<enumeration name=\"' + name.encode('utf-8') + '\">'
        xml += ' '.join('<value>' + v + '</value>' for v in values)
        xml += '</enumeration>'
        return self._reqXml('PUT', '/admin/customfield/bundle', body=xml.encode('utf8'), ignoreStatus=400)

    def addValueToEnumBundle(self, name, value):
        return self.addValueToBundle(self.getEnumBundle(name), value)

    def addValuesToEnumBundle(self, name, values):
        return ", ".join(self.addValueToEnumBundle(name, value) for value in values)

    def getYouTrackBuildNumber(self):
        response, content = self._req('GET',
                                      self.url + '/api/config?fields=build',
                                      ignoreStatus=404,
                                      content_type='application/json')
        if response.status != 200 or not content:
            return 0
        try:
            return int(json.loads(content).get('build', 0))
        except ValueError:
            return 0

    def isMarkdownSupported(self):
        return self.getYouTrackBuildNumber() > 39406

    bundle_paths = {
        "enum": "bundle",
        "build": "buildBundle",
        "ownedField": "ownedFieldBundle",
        "state": "stateBundle",
        "version": "versionBundle",
        "user": "userBundle"
    }

    bundle_types = {
        "enum": lambda xml, yt: youtrack.EnumBundle(xml, yt),
        "build": lambda xml, yt: youtrack.BuildBundle(xml, yt),
        "ownedField": lambda xml, yt: youtrack.OwnedFieldBundle(xml, yt),
        "state": lambda xml, yt: youtrack.StateBundle(xml, yt),
        "version": lambda xml, yt: youtrack.VersionBundle(xml, yt),
        "user": lambda xml, yt: youtrack.UserBundle(xml, yt)
    }



import copy

class LinkImporter(object):
    def __init__(self, target, project_id=None, query=None):
        self.target = target
        self.created_issue_ids = self._get_all_issue_ids_set(self.target, project_id, query) if project_id else set([])
        self.links = []
        self.verbose_mode = False
        self.target_name = "youtrack"
        self.header = "[Link importer]"

    def setVerboseMode(self, mode):
        self.verbose_mode = mode

    def resetConnections(self, target):
        self.target = target

    def setYoutrackName(self, name):
        self.target_name = name

    def setLogHeader(self, header):
        self.header = header

    def _get_all_issue_ids_set(self, yt, project_id, query):
        if not query: query = ''
        start = 0
        batch = 50
        result = set([])
        while True:
            issues = yt.getIssues(project_id, query, start, batch)
            if not len(issues): break
            for issue in issues:
                result.add(issue.id)
            start += batch
        return result

    def resetAvailableIssues(self):
        self.created_issue_ids = set([])

    def addAvailableIssuesFrom(self, project_id):
        self.created_issue_ids |= self._get_all_issue_ids_set(self.target, project_id, None) if project_id else set([])

    def addAvailableIssues(self, issues):
        self.created_issue_ids |= set([issue.id for issue in issues])

    def addAvailableIssue(self, issue):
        self.created_issue_ids.add(issue.id)

    def collectLinks(self, links):
        self.links += links

    def importLinks(self, links):
        maxLinks = 100
        links_to_import = []
        for link in links:
            if link.target not in self.created_issue_ids:
                print(self.header + ' failed to import link ' + self._getPrettyLink(link) + ' to ' + self.target_name + ' because ' + link.target + ' was not imported')
            elif link.source not in self.created_issue_ids:
                print(self.header + 'failed to import link ' + self._getPrettyLink(link) + ' to ' + self.target_name + ' because ' + link.source + ' was not imported')
            else:
                links_to_import.append(link)
                if len(links_to_import) == maxLinks:
                    self._import_links_batch(links_to_import)
                    links_to_import = []
        if len(links_to_import):
            self._import_links_batch(links_to_import)

    def _import_links_batch(self, links_to_import):
        if not self.verbose_mode:
            self.target.importLinks(links_to_import)
        for link in links_to_import:
            print(self.header + ' imported ' + self._getPrettyLink(link) + ' to ' + self.target_name)

    def importCollectedLinks(self):
        self.importLinks(self.links)
        self.links = []

    def _getPrettyLink(self, link):
        return link.typeName + ' link: ' + link.source + '->' + link.target

    def checkLink(self, link):
        return link.source in self.created_issue_ids and link.target in self.created_issue_ids

class LinkSynchronizer(object):
    def __init__(self, master_executor, slave_executor, binder):
        self.issue_binder = binder
        self.slaveExecutor = slave_executor
        self.masterExecutor = master_executor
        self.master_links = []
        self.slave_links = []

    def collectLinksToSyncById(self, master_issue_id, slave_issue_id):

        slave_links = self.slaveExecutor.yt.getLinks(slave_issue_id, True) if slave_issue_id else []
        master_links = self.masterExecutor.yt.getLinks(master_issue_id, True) if master_issue_id else []

        to_master_links = set([self._convertSlaveLinkForMaster(link) for link in slave_links if self.check_slave_link(link)]) - set(master_links)
        to_slave_links = set([self._convertMasterLinkForSlave(link) for link in master_links if self.check_master_link(link)]) - set(slave_links)

        self.master_links += to_master_links
        self.slave_links += to_slave_links

    def _convertSlaveLinkForMaster(self, slave_link):
        link_copy = copy.copy(slave_link)
        link_copy.source = self.issue_binder.slaveIssueIdToMasterIssueId(slave_link.source)
        link_copy.target = self.issue_binder.slaveIssueIdToMasterIssueId(slave_link.target)
        return link_copy

    def _convertMasterLinkForSlave(self, master_link):
        link_copy = copy.copy(master_link)
        link_copy.source = self.issue_binder.masterIssueIdToSlaveIssueId(master_link.source)
        link_copy.target = self.issue_binder.masterIssueIdToSlaveIssueId(master_link.target)
        return link_copy

    def check_slave_link(self, link):
        return self.issue_binder.checkSlaveId(link.source) and self.issue_binder.checkSlaveId(link.target)

    def check_master_link(self, link):
        return self.issue_binder.checkMasterId(link.source) and self.issue_binder.checkMasterId(link.target)

    def syncCollectedLinks(self):
        self.slaveExecutor.importLinks(self.slave_links, self.issue_binder.getPermittedSlaveIds())
        self.masterExecutor.importLinks(self.master_links, self.issue_binder.getPermittedMasterIds())
        self.slave_links = []
        self.master_links = []

class IssueBinder(object):
    def __init__(self, s_to_m):
        self.s_to_m = copy.copy(s_to_m)
        self.m_to_s = {}
        for s_id , m_id in list(s_to_m.items()):
            self.m_to_s[m_id] = s_id

    def slaveIssueIdToMasterIssueId(self, slave_issue_id):
        return str(self.s_to_m[str(slave_issue_id)])

    def masterIssueIdToSlaveIssueId(self, master_issue_id):
        return str(self.m_to_s[str(master_issue_id)])

    def addBinding(self, master_id, slave_id):
        self.s_to_m[slave_id] = master_id
        self.m_to_s[master_id] = slave_id

    def getPermittedMasterIds(self):
        return list(self.m_to_s.keys())

    def getPermittedSlaveIds(self):
        return list(self.s_to_m.keys())

    def checkSlaveId(self, id):
        return id in self.s_to_m

    def checkMasterId(self, id):
        return id in self.m_to_s


from youtrack import YouTrackException

COMPARISON_LENGTH = 10
COMMAND = 'comment'

class CommentSynchronizer(object):
    def __init__(self, master, slave, master_executor, slave_executor):
        self.master = master
        self.slave = slave
        self.executors = {master : master_executor, slave : slave_executor}

    def syncComments(self, master_id, slave_id):
        slave_comments = self.slave.getComments(slave_id)
        master_comments = self.master.getComments(master_id)
        if len(slave_comments) or len(master_comments):
            master_texts = set([cm.text[0:COMPARISON_LENGTH] for cm in master_comments])
            slave_texts = set([cm.text[0:COMPARISON_LENGTH] for cm in slave_comments])
            slave_unique = [cm for cm in slave_comments if cm.text[0:COMPARISON_LENGTH] not in master_texts]
            master_unique = [cm for cm in master_comments if cm.text[0:COMPARISON_LENGTH] not in slave_texts]
            for cm in slave_unique:
                self._sync_comment(self.master, self.slave, master_id, cm.text, cm.author)
            for cm in master_unique:
                self._sync_comment(self.slave, self.master, slave_id, cm.text, cm.author)

    def _sync_comment(self, to_yt, from_yt, issue_id, comment_text, run_as):
        if comment_text is not None and comment_text != '':
            self._try_to_sync_user(to_yt, from_yt, run_as)
            self.executors[to_yt].executeCommand(issue_id, COMMAND, comment=comment_text, run_as=run_as)

    def _try_to_sync_user(self, to_yt, from_yt, login):
        try:
            to_yt.getUser(login)
        except YouTrackException:
            user_to_import = from_yt.getUser(login)
            self.executors[to_yt].executeUserImport(user_to_import)



from youtrack import YouTrackException


def utf8encode(source):
    if isinstance(source, str):
        source = source.encode('utf-8')
    return source


def _create_custom_field_prototype(connection, cf_type, cf_name, auto_attached=False, additional_params=dict([])):
    field = _get_custom_field(connection, cf_name)
    if field is not None:
        if field.type != cf_type:
            msg = "Custom field with name [ %s ] already exists. It has type [ %s ] instead of [ %s ]" %\
                  (utf8encode(cf_name), field.type, cf_type)
            raise LogicException(msg)
    else:
        connection.createCustomFieldDetailed(cf_name, cf_type, False, True, auto_attached, additional_params)


def _get_custom_field(connection, cf_name):
    existing_fields = [item for item in connection.getCustomFields() if utf8encode(item.name).lower() ==
                                                                        utf8encode(cf_name).lower()]
    if len(existing_fields):
        return existing_fields[0]
    return None

def create_custom_field(connection, cf_type, cf_name, auto_attached, value_names=None, bundle_policy="0"):
    """
    Creates custom field prototype(if not exist) and sets default values bundle if needed

     Args:
        connection: An opened Connection instance.
        cf_type: Type of custom field to be created
        cf_name: Name of custom field that should be created (if not exists)
        auto_attached: If this field should be auto attached or not.
        value_names: Values, that should be attached with this cf by default. If None, no bundle is created to this field
            if empty, empty bundle is created.
    Raises:
        LogicException: If custom field already exists, but has wrong type.
        YouTrackException: If something is wrong with queries.
    """
    if (value_names is None) and (not auto_attached or "[" not in cf_type):
        _create_custom_field_prototype(connection, cf_type, cf_name, auto_attached)
        return
    if value_names is None:
        value_names = set([])
    else:
        value_names = set(value_names)
    field = _get_custom_field(connection, cf_name)
    bundle = None
    if field is not None:
        if hasattr(field, "defaultBundle"):
            bundle = connection.getBundle(field.type, field.defaultBundle)
        elif field.autoAttached:
            return 
        else:
            bundle = create_bundle_safe(connection, cf_name + "_bundle", cf_type)
    else:
        bundle = create_bundle_safe(connection, cf_name + "_bundle", cf_type)
        _create_custom_field_prototype(connection, cf_type, cf_name, auto_attached,
                {"defaultBundle": bundle.name,
                 "attachBundlePolicy": bundle_policy})
    for value_name in value_names:
        try:
            connection.addValueToBundle(bundle, value_name)
        except YouTrackException:
            pass
#
#    values_to_add = calculate_missing_value_names(bundle, value_names)
#    [connection.addValueToBundle(bundle, name) for name in values_to_add]


#    if field is None:
#        bundle_name = cf_name + "_bundle"
#        _create_bundle_safe(connection, bundle_name, cf_type)
#        bundle = connection.getBundle(cf_type, bundle_name)
#        values_to_add = calculate_missing_value_names(bundle, value_names)
#
#
#        for value in values_to_add:
#            connection.addValueToBundle(bundle, value)
#
#


def process_custom_field(connection, project_id, cf_type, cf_name, value_names=None):
    """
    Creates custom field and attaches it to the project. If custom field already exists and has type
    cf_type it is attached to the project. If it has another type, LogicException is raised. If project field already
    exists, uses it and bundle from it. If not, creates project field and bundle with name <cf_name>_bundle_<project_id> for it.
    Adds value_names to bundle.
    Args:
        connection: An opened Connection instance.
        project_id: Id of the project to attach CF to.
        cf_type: Type of cf to be created.
        cf_name: Name of cf that should be created (if not exists) and attached to the project (if not yet attached)
        value_names: Values, that cf must have. If None, does not create any bundle for the field. If empty list, creates bundle, but
            does not create any value_names in it. If bundle already contains some value_names, only value_names that do not already exist are added.

    Raises:
        LogicException: If custom field already exists, but has wrong type.
        YouTrackException: If something is wrong with queries.  

    """
    _create_custom_field_prototype(connection, cf_type, cf_name)
    if cf_type[0:-3] not in connection.bundle_types:
        value_names = None
    elif value_names is None:
        value_names = []

    existing_project_fields = [item for item in connection.getProjectCustomFields(project_id) if
                               utf8encode(item.name) == cf_name]
    values_to_add = []
    bundle = None
    if len(existing_project_fields):
        if value_names is None:
            return
        bundle = connection.getBundle(cf_type, existing_project_fields[0].bundle)
        values_to_add = calculate_missing_value_names(bundle, value_names)
    else:
        if value_names is None:
            connection.createProjectCustomFieldDetailed(project_id, cf_name, "No " + cf_name)
            return
        bundle = create_bundle_safe(connection, cf_name + "_bundle_" + project_id, cf_type)
        values_to_add = calculate_missing_value_names(bundle, value_names)
        connection.createProjectCustomFieldDetailed(project_id, cf_name, "No " + cf_name,
                                                    params={"bundle": bundle.name})
    [connection.addValueToBundle(bundle, bundle.createElement(name)) for name in values_to_add]


def add_values_to_bundle_safe(connection, bundle, values):
    """
    Adds values to specified bundle. Checks, whether each value already contains in bundle. If yes, it is not added.

    Args:
        connection: An opened Connection instance.
        bundle: Bundle instance to add values in.
        values: Values, that should be added in bundle.

    Raises:
        YouTrackException: if something is wrong with queries.
    """
    for value in values:
        try:
            connection.addValueToBundle(bundle, value)
        except YouTrackException as e:
            if e.response.status == 409:
                print("Value with name [ %s ] already exists in bundle [ %s ]" % \
                      (utf8encode(value.name), utf8encode(bundle.name)))
            else:
                raise e


def create_bundle_safe(connection, bundle_name, bundle_type):
    bundle = connection.bundle_types[bundle_type[0:-3]](None, None)
    bundle.name = bundle_name
    try:
        connection.createBundle(bundle)
    except YouTrackException as e:
        if e.response.status == 409:
            print("Bundle with name [ %s ] already exists" % bundle_name)
        else:
            raise e
    return connection.getBundle(bundle_type, bundle_name)


def calculate_missing_value_names(bundle, value_names):
    bundle_elements_names = [elem.name.lower() for elem in bundle.values]
    return [value for value in value_names if value.lower() not in bundle_elements_names]


class LogicException(Exception):
    def __init__(self, msg):
        Exception.__init__(self, msg)



import youtrack

PROHIBITED = '/'

def utf8encode(source):
    if isinstance(source, str):
        source = source.encode('utf-8')
    return source


class UserImporter(object):
    def __init__(self, source, target, caching_users=True, import_groups=True):
        self.source = source
        self.target = target
        self.caching_users = caching_users
        self.import_groups = import_groups
        #self.created_user_logins = set([user.login for user in target.getUsers()]) if caching_users else set([])
        self.created_user_logins = set([])
        self.created_group_names = set([group.name for group in target.getGroups()])
        self.created_role_names = set([role.name for role in target.getRoles()])
        self.created_project_ids = set(target.getProjectIds())

    def addCreatedProjects(self, project_ids):
        self.created_project_ids |= set(project_ids)

    def resetConnections(self, source, target):
        self.source = source
        self.target = target

    def importUser(self, user):
        filtered_user = self._filter_user(user)
        if filtered_user:
            self.target.importUsers([filtered_user])

    def importUsersRecursively(self, users):
        total_users = len(users)
        if not total_users: return
        user_list = list(users)
        max_users = 100
        start = 0
        imported_size = 0
        while start < total_users:
            user_batch = user_list[start:start + max_users]
            imported_size += self._import_user_batch_recursively(user_batch)
            start += max_users
        return imported_size

    def _import_groups_of(self, yt_user):
        user_groups = self.source.getUserGroups(yt_user.login)
        for group in user_groups:
            if group.name not in self.created_group_names:
                try:
                    self.createGroup(group)
                except Exception as ex:
                    print(utf8encode(repr(ex)))
                    continue
            print("Set " + utf8encode(yt_user.login) + " to " + utf8encode(group.name))
            self.target.setUserGroup(yt_user.login, group.name)

    def _import_user_batch_recursively(self, users):
        if not len(users): return 0
        users_to_import = []
        for user in users:
            filtered_user = self._filter_user(user)
            if filtered_user: users_to_import.append(filtered_user)
        self.target.importUsers(users_to_import)
        for yt_user in users_to_import:
            if self.import_groups:
                self._import_groups_of(yt_user)
            if self.caching_users: self.created_user_logins.add(yt_user.login)
        return len(users_to_import)

    def _filter_user(self, user):
        if (not self.caching_users or user.login not in self.created_user_logins) and self._check_login(user.login):
            if not hasattr(user, "email"):
                user.email = "<no email>"
            return user
        return None

    def importGroupsWithoutUsers(self, groups):
        if not len(groups): return
        for group in groups:
            if group.name not in self.created_group_names:
                try:
                    self.createGroup(group)
                except Exception as ex:
                    print(utf8encode(repr(ex)))

    def createGroup(self, group):
        group_roles = self.source.getGroupRoles(group.name)
        self.target.createGroup(group)
        self.created_group_names.add(group.name)
        for user_role in group_roles:
            role = self.source.getRole(user_role.name)
            if role.name not in self.created_role_names:
                self._create_role(role)
            self._add_user_role_to_group_safely(group, user_role)


    def _add_user_role_to_group_safely(self, group, user_role):
        restricted_user_role = youtrack.UserRole()
        restricted_user_role.name = user_role.name
        for project in user_role.projects:
            if project in self.created_project_ids:
                restricted_user_role.projects.append(project)
        self.target.addUserRoleToGroup(group, restricted_user_role)

    def _create_role(self, role):
        permissions = self.source.getRolePermissions(role)
        self.target.createRole(role)
        self.created_role_names.add(role.name)
        for prm in permissions:
            self.target.addPermissionToRole(role, prm)

    def _check_login(self, login):
        failed = 1 in [c in login for c in PROHIBITED]
        if failed: print("Could not import user [" + login + "], login contains prohibited chars: " + PROHIBITED)
        return not failed



from datetime import datetime

LOGGING = True
log_file_name_format = 'log_%d_%m_%Y'
error_file_name_format = 'error_%d_%m_%Y'
MASTER_NAME = "Master"
SLAVE_NAME = "Slave"
UNDEFINED = "Undefined"

class Logger(object):
    def __init__(self, master, slave, master_root_login, slave_root_login):
        today = datetime.now()
        self.log_file = open(today.strftime(log_file_name_format), 'w')
        self.error_file = open(today.strftime(error_file_name_format), 'w')
        self.master = master
        self.slave = slave
        self.master_root_login = master_root_login
        self.slave_root_login = slave_root_login


    def logAction(self, action_name, yt, message, run_as=None):
        line = self._prepare_line(action_name, yt, message, run_as)
        print(line)
        try:
            if LOGGING: self.log_file.write(str(line) + '\n')
        except UnicodeError as e:
            print(e)

    def logError(self, error, action_name, yt, message, run_as=None):
        if LOGGING:
            line = self._prepare_line(action_name, yt, message, run_as)
            try:
                self.log_file.write(str(line) + '\n')
                self.error_file.write(str(line) + '\n')
                self.error_file.write(str(error) + '\n')
                self.error_file.write('---------------------------------------------------------\n')
                print(line)
                print(error)
            except UnicodeError as e:
                print(e)

    def finalize(self):
        self.log_file.close()
        self.error_file.close()

    def _prepare_line(self, action_name, yt, message, run_as):
        yt_name = UNDEFINED if None else MASTER_NAME if yt == self.master else SLAVE_NAME
        user_login = (self.master_root_login if yt == self.master else self.slave_root_login) if run_as is None else run_as
        return '[Sync, ' + action_name + ' in ' + yt_name + '] ' + message + ' on behalf of ' + user_login



def import_distinct_roles(source, target):
    source_group_names = set([group.name for group in source.getGroups()])
    required_target_group_names = [group.name for group in target.getGroups() if group.name in source_group_names]

    required_for_import_role_names = set([])
    for name in required_target_group_names:
        required_for_import_role_names |= set([ur.name for ur in source.getGroupRoles(name)])

    roles = source.getRoles()
    slave_role_names = set([role.name for role in target.getRoles()])
    for role in roles:
        if role.name not in slave_role_names:
            if role.name in required_for_import_role_names:
                target.createRole(role)
                permissions = source.getRolePermissions(role)
                for permission in permissions:
                    target.addPermissionToRole(role, permission)



import itertools

import youtrack
from youtrack import YouTrackException, Issue
from youtrack.importHelper import create_custom_field

__author__ = 'user'

NAME = 'name'
TYPE = 'type'
POLICY = 'bundle_policy'
AUTO_ATTACHED = 'auto_attached'
NUMBER_IN_PROJECT = 'numberInProject'


class YouTrackImporter(object):
    def __init__(self, source, target, import_config):
        self._source = source
        self._target = target
        self._import_config = import_config

    def do_import(self, projects, new_projects_owner_login='root'):
        project_ids = list(projects.keys())
        self._create_auto_attached_fields()
        self._create_custom_fields(project_ids)
        for project_id, project_name in list(projects.items()):
            self._create_project(project_id, project_name, new_projects_owner_login)
            self._attach_fields_to_project(project_id)
            self._add_value_to_fields_in_project(project_id)
            self._import_issues(project_id)
        self._import_tags(project_ids)
        self._import_issue_links(project_ids)

    def _create_auto_attached_fields(self):
        predefined_fields = self._import_config.get_predefined_fields()
        for field in predefined_fields:
            self._create_field(field[NAME], field[TYPE], field.get(POLICY))

    def _create_field(self, field_name, field_type, attach_bundle_policy, auto_attached=True):
        if field_name in youtrack.EXISTING_FIELDS:
            return
        create_custom_field(self._target, field_type, field_name, auto_attached, bundle_policy=attach_bundle_policy)

    def _create_custom_fields(self, project_ids):
        custom_fields = self._get_custom_fields_for_projects(project_ids)
        for yt_field in custom_fields:
            if (yt_field is not None) and (yt_field[TYPE] is not None):
                self._create_field(yt_field[NAME], yt_field[TYPE], yt_field.get(POLICY), yt_field[AUTO_ATTACHED])

    def _get_custom_fields_for_projects(self, project_ids):
        raise NotImplementedError()

    def _create_project(self, project_id, project_name, project_lead_login):
        try:
            self._target.getProject(project_id)
        except YouTrackException:
            self._target.createProjectDetailed(project_id, project_name, '', project_lead_login)

    def _attach_fields_to_project(self, project_id):
        for yt_field in self._get_custom_fields_for_projects([project_id]):
            field_name = yt_field[NAME]
            try:
                self._target.createProjectCustomFieldDetailed(project_id, field_name, 'No ' + field_name)
            except YouTrackException:
                pass
                # print(u'Field [%s] is already attached' % field_name)

    def _import_issues(self, project_id):
        limit = 100
        all_issues = self._get_issues(project_id)
        while True:
            issues = list(itertools.islice(all_issues, None, limit))
            if not len(issues):
                break
            self._target.importIssues(project_id, project_id + ' assignees',
                [self._to_yt_issue(issue, project_id) for issue in issues])
            for issue in issues:
                issue_id = self._get_issue_id(issue)
                issue_attachments = self._get_attachments(issue)
                yt_issue_id = '%s-%s' % (project_id, issue_id)
                self._import_attachments(yt_issue_id, issue_attachments)

    def _import_tags(self, project_ids):
        limit = 100
        existing_tags = set([])
        for project_id in project_ids:
            all_tags = self._get_issue_tags(project_id)
            while True:
                l = list(itertools.islice(all_tags, None, limit))
                if not len(l):
                    break
                issue_tags = list(zip(*l))[1]
                existing_tags |= set([item for tags in issue_tags for item in tags])
        self._do_import_tags(project_ids, existing_tags)

    def _is_prefix_of_any_other_tag(self, tag, other_tags):
        for t in other_tags:
            if t.startswith(tag) and (t != tag):
                return True
        return False


    def _do_import_tags(self, project_ids, collected_tags):
        tags_to_import_now = set([])
        tags_to_import_after = set([])
        for tag in collected_tags:
            if self._is_prefix_of_any_other_tag(tag, collected_tags):
                tags_to_import_after.add(tag)
            else:
                tags_to_import_now.add(tag)
        for project_id in project_ids:
            for (issue_id, tags) in self._get_issue_tags(project_id):
                yt_issue_id = '%s-%s' % (project_id, issue_id)
                for tag in tags:
                    if tag in tags_to_import_now:
                        try:
                            self._target.executeCommand(yt_issue_id, 'tag ' + tag)
                        except YouTrackException:
                            print(('Failed to import tag for issue [%s]' % yt_issue_id))
        if len(tags_to_import_after):
            self._do_import_tags(project_ids, tags_to_import_after)

    def _import_issue_links(self, project_ids):
        limit = 100
        for project_id in project_ids:
            after = 0
            while True:
                links = self._get_issue_links(project_id, after, limit)
                if not len(links):
                    break
                self._target.importLinks(links)
                after += limit

    def process_field(self, key, project_id, result, value):
        # we do not need fields with empty values
        if value is None:
            return
        if isinstance(value, list) and not len(value):
            return
        if (isinstance(value, str) or isinstance(value, str)) and not len(value):
            return

        #get yt field name and field type
        field_name = self._get_field_name(key, project_id)
        if field_name is None or field_name == NUMBER_IN_PROJECT:
            return
        field_type = self._get_field_type(field_name)
        if (field_type is None) and (field_name not in youtrack.EXISTING_FIELDS):
            return
        value = self.get_field_value(field_name, field_type, value)
        if isinstance(value, list):
            for v in value:
                self._add_value_to_field(project_id, field_name, field_type, v)
        else:
            self._add_value_to_field(project_id, field_name, field_type, value)
        if (field_type is not None) and field_type.startswith('user'):
            if isinstance(value, list):
                value = [v.login for v in value]
            else:
                value = value.login
        if not isinstance(value, list):
            value = str(value)
        result[field_name] = value

    def _to_yt_issue(self, issue, project_id):
        result = Issue()
        result.comments = [self._to_yt_comment(comment) for comment in self._get_comments(issue)]
        result.numberInProject = self._get_issue_id(issue)
        for (key, value) in list(issue.items()):
            self.process_field(key, project_id, result, value)
        return result

    def _get_field_name(self, field_name, project_id):
        field_name = self._import_config.get_field_name(field_name)
        if field_name in youtrack.EXISTING_FIELDS:
            return field_name
        try:
            self._target.getProjectCustomField(project_id, field_name)
            return field_name
        except YouTrackException:
            return None

    def _get_field_type(self, field_name):
        if field_name in youtrack.EXISTING_FIELD_TYPES:
            return youtrack.EXISTING_FIELD_TYPES[field_name]
        try:
            return self._target.getCustomField(field_name).type
        except YouTrackException:
            return None

    def _import_user(self, user):
        self._target.importUsers([user])
        for group in user.getGroups():
            try:
                self._target.createGroup(group)
            except YouTrackException:
                pass
            self._target.setUserGroup(user.login, group.name)

    def _add_value_to_field(self, project_id, field_name, field_type, value):
        if (field_type is not None) and field_type.startswith('user'):
            self._import_user(value)
            value = value.login
        if field_name in youtrack.EXISTING_FIELDS:
            return
        custom_field = self._target.getProjectCustomField(project_id, field_name)
        if hasattr(custom_field, 'bundle'):
            bundle = self._target.getBundle(field_type, custom_field.bundle)
            try:
                self._target.addValueToBundle(bundle, value)
            except YouTrackException:
                pass

    def get_field_value(self, field_name, field_type, value):
        if value is None:
            return None
        values_map = self._import_config.get_value_mapping(field_name)
        if isinstance(value, list):
            return [self.get_field_value(field_name, field_type, v) for v in value]
        if field_type.startswith('user'):
            return self._to_yt_user(value)
        if field_type.lower() == "date":
            return self.to_unix_date(value)
        if isinstance(value, str):
            return values_map.get(value, value)
        if isinstance(value, int):
            return values_map.get(value, str(value))

    def to_unix_date(self, date):
        return date

    def _add_value_to_fields_in_project(self, project_id):
        for field in self._get_fields_with_values(project_id):
            field_name = self._get_field_name(field[NAME], project_id)
            pcf = self._target.getProjectCustomField(project_id, field_name)
            if hasattr(pcf, 'bundle'):
                field_type = pcf.type[0:-3]
                bundle = self._target.getBundle(field_type, pcf.bundle)
                yt_values = [v for v in [field['converter'](value, bundle,
                    lambda name, value_name: self._import_config.get_field_value(name, field_type, value_name)) for
                                         value in
                                         field['values']] if len(v)]
                for value in yt_values:
                    self._target.addValueToBundle(bundle, value)

    def _get_issue_id(self, issue):
        return str(issue[self._import_config.get_key_for_field_name(NUMBER_IN_PROJECT)])

    #Following method should be implemented in inheritors:

    def _get_fields_with_values(self, project_id):
        return []

    def _to_yt_comment(self, comment):
        raise NotImplementedError

    def _get_attachments(self, issue_id):
        return []

    def _get_issues(self, project_id):
        raise NotImplementedError

    def _import_attachments(self, issue_id, issue_attachments):
        for attach in issue_attachments:
            self._target.createAttachmentFromAttachment(issue_id, attach)

    def _get_comments(self, issue):
        raise NotImplementedError

    def _get_issue_tags(self, project_id):
        key = self._import_config.get_key_for_field_name('Tags')
        return ((self._get_issue_id(issue), issue[key]) for issue in self._get_issues(project_id) if (key in issue ) and len(issue[key]))

    def _get_issue_links(self, project_id, after, limit):
        return []

    def _to_yt_user(self, value):
        raise NotImplementedError


class YouTrackImportConfig(object):
    def __init__(self, name_mapping, type_mapping, value_mapping=None, link_type_mapping=None):
        self._name_mapping = name_mapping
        self._type_mapping = type_mapping
        self._value_mapping = value_mapping if value_mapping is not None else {}
        self._link_type_mapping = link_type_mapping if link_type_mapping is not None else {}

    def _get_default_auto_attached(self):
        return True

    def _get_default_bundle_policy(self):
        return 0

    def get_field_name(self, field_name):
        return field_name if field_name not in self._name_mapping else self._name_mapping[field_name]

    def get_predefined_fields(self):
        return []

    def get_link_type(self, type):
        return self._link_type_mapping[type] if type in self._link_type_mapping else type

    def get_key_for_field_name(self, field_name):
        for (key, value) in list(self._name_mapping.items()):
            if value == field_name:
                return key
        return field_name

    def get_value_mapping(self, field_name):
        return self._value_mapping[field_name] if field_name in self._value_mapping else {}
